@@ -1,3 +1,12 @@
+//
+// IconView.cs:
+//
+// Authors:
+//    Ettore Perazzoli
+//    Larry Ewing <lewing@novell.com>
+//
+// (C) 2003 Novell, Inc.
+//
 using Gtk;
 using Gdk;
 using Gnome;
@@ -55,6 +64,17 @@ public class IconView : Gtk.Layout {
 		}
 	}
 
+	private bool display_dates = true;
+	public bool DisplayDates {
+		get {
+			return display_dates;
+		}
+
+		set {
+			display_dates = value;
+			QueueResize ();
+		}
+	}
 
 	// Size of the frame around the thumbnail.
 	private const int CELL_BORDER_WIDTH = 10;
@@ -63,10 +83,10 @@ public class IconView : Gtk.Layout {
 	private const int BORDER_SIZE = 6;
 
 	// Thickness of the outline used to indicate selected items. 
-	private const int SELECTION_THICKNESS = 3;
+	private const int SELECTION_THICKNESS = 5;
 
 	// Size of the tag icon in the view.
-	private const int TAG_ICON_SIZE = 12;
+	private const int TAG_ICON_SIZE = 16;
 
 	// Horizontal spacing between the tag icons
 	private const int TAG_ICON_HSPACING = 2;
@@ -82,9 +102,6 @@ public class IconView : Gtk.Layout {
 	private int cell_width;
 	private int cell_height;
 
-	// Query we are displaying.
-	private PhotoQuery query;
-
 	// The first pixel line that is currently on the screen (i.e. in the current
 	// scroll region).  Used to compute the area that went offscreen in the "changed"
 	// signal handler for the vertical GtkAdjustment.
@@ -94,12 +111,22 @@ public class IconView : Gtk.Layout {
 	private Hashtable selected_cells;
 
 	// Drag and drop bookkeeping. 
-	private bool in_drag;
 	private int click_x, click_y;
 
 	// Focus Handling
-	private int focus_cell = 0;
-
+	private int real_focus_cell;
+	public int FocusCell {
+		set {
+			if (value != real_focus_cell) {
+				InvalidateCell (value);
+				InvalidateCell (real_focus_cell);
+				real_focus_cell = value;
+			}
+		}
+		get {
+			return real_focus_cell;
+		}
+	}
 	// Number of consecutive GDK_BUTTON_PRESS on the same cell, to
 	// distinguish the GDK_2BUTTON_PRESS events that we actually care
 	// about.
@@ -108,9 +135,7 @@ public class IconView : Gtk.Layout {
 	// The pixbuf we use when we can't load a thumbnail.
 	static Pixbuf error_pixbuf;
 
-
 	// Public events.
-
 	public delegate void DoubleClickedHandler (IconView view, int clicked_item);
 	public event DoubleClickedHandler DoubleClicked;
 
@@ -122,15 +147,12 @@ public class IconView : Gtk.Layout {
 
 	public IconView () : base (null, null)
 	{
-		pixbuf_loader = new PixbufLoader (ThumbnailWidth);
+		pixbuf_loader = new PixbufLoader (256);
 		pixbuf_loader.OnPixbufLoaded += new PixbufLoader.PixbufLoadedHandler (HandlePixbufLoaded);
 
 		selected_cells = new Hashtable ();
 
 		ScrollAdjustmentsSet += new ScrollAdjustmentsSetHandler (HandleScrollAdjustmentsSet);
-		SizeAllocated += new SizeAllocatedHandler (HandleSizeAllocated);
-		ExposeEvent += new ExposeEventHandler (HandleExposeEvent);
-		MotionNotifyEvent += new MotionNotifyEventHandler (HandleMotionNotifyEvent);
 		
 		ButtonPressEvent += new ButtonPressEventHandler (HandleButtonPressEvent);
 		ButtonReleaseEvent += new ButtonReleaseEventHandler (HandleButtonReleaseEvent);
@@ -138,38 +160,39 @@ public class IconView : Gtk.Layout {
 
 		DestroyEvent += new DestroyEventHandler (HandleDestroyEvent);
 
-		AddEvents ((int) EventMask.KeyPressMask);
+		AddEvents ((int) EventMask.KeyPressMask
+			   | (int) EventMask.KeyReleaseMask 
+			   | (int) EventMask.PointerMotionMask);
 		
 		CanFocus = true;
+
+		Gdk.Color color = this.Style.Background (Gtk.StateType.Normal);
+		color.Red = (ushort) (color.Red / 2);
+		color.Blue = (ushort) (color.Blue / 2);
+		color.Green = (ushort) (color.Green / 2);
+		ModifyBg (Gtk.StateType.Normal, color);
+	}
 	
-		string [] types = new string [1];
-		types [0] = "text/uri-list";
-		GtkDnd.SetAsDestination (this, types);
-		DragDrop += new DragDropHandler (HandleDragDrop);
-	}
-
-	private void OnReload (PhotoQuery query)
+	public IconView (FSpot.IPhotoCollection collection) : this () 
 	{
-		// FIXME we should probably try to merge the selection forward
-		// but it needs some thought to be efficient.
-		UnselectAllCells ();
-		QueueResize ();
+		this.collection = collection;
 	}
+	
 
-	public IconView (PhotoQuery query) : this ()
-	{
-		this.query = query;
-		query.Reload += new PhotoQuery.ReloadHandler (OnReload);
-	}
+	protected IconView (IntPtr raw) : base (raw) {}
 
+	//
+	// IPhotoSelection
+	//
 
-	public PhotoQuery Query {
+	protected FSpot.IPhotoCollection collection;
+	public FSpot.IPhotoCollection Collection {
 		get {
-			return query;
+			return collection;
 		}
 	}
 
-	public int [] Selection {
+	public int [] SelectedIdxs {
 		get {
 			int [] selection = new int [selected_cells.Count];
 
@@ -182,48 +205,82 @@ public class IconView : Gtk.Layout {
 		}
 	}
 
+	public int SelectedIdxCount 
+	{
+		get {
+			return selected_cells.Count;
+		}
+	}
+	
+	public bool IdxIsSelected (int num)
+	{
+		return CellIsSelected (num);
+	}
+
+	public int CurrentIdx {
+		get {
+			if (selected_cells.Count == 1 && IdxIsSelected (FocusCell))
+				return FocusCell;
+			else 
+				return -1;
+		}
+	}
 
 	// Updating.
 
 	public void UpdateThumbnail (int thumbnail_num)
 	{
-		Photo photo = query.Photos [thumbnail_num];
+		Photo photo = collection.Photos [thumbnail_num];
 		string thumbnail_path = Thumbnail.PathForUri ("file://" + photo.DefaultVersionPath, ThumbnailSize.Large);
+
+		//Console.WriteLine ("remove {0}", thumbnail_path);
 		ThumbnailCache.Default.RemoveThumbnailForPath (thumbnail_path);
 		InvalidateCell (thumbnail_num);
 	}
 
 
-	// Private utility methods.
+	// Cell Geometry
 
-	static private Pixbuf ErrorPixbuf ()
+	public int CellAtPosition (int x, int y)
 	{
-		if (IconView.error_pixbuf == null)
-			IconView.error_pixbuf = PixbufUtils.LoadFromAssembly ("f-spot-question-mark.png");
-
-		return IconView.error_pixbuf;
-	}
-
-	private int CellAtPosition (int x, int y)
-	{
-		if (query == null)
+		if (collection == null)
 			return -1;
 
 		if (x < BORDER_SIZE || x >= BORDER_SIZE + cells_per_row * cell_width)
 			return -1;
-		if (y < BORDER_SIZE || y >= BORDER_SIZE + (query.Photos.Length / cells_per_row + 1) * cell_height)
+		if (y < BORDER_SIZE || y >= BORDER_SIZE + (collection.Photos.Length / cells_per_row + 1) * cell_height)
 			return -1;
 
 		int column = (int) ((x - BORDER_SIZE) / cell_width);
 		int row = (int) ((y - BORDER_SIZE) / cell_height);
 		int cell_num = column + row * cells_per_row;
 
-		if (cell_num < query.Photos.Length)
+		if (cell_num < collection.Photos.Length)
 			return (int) cell_num;
 		else
 			return -1;
 	}
 
+	public void GetCellCenter (int cell_num, out int x, out int y)
+	{
+		if (cell_num == -1) {
+			x = -1;
+			y = -1;
+		}
+
+		x = BORDER_SIZE + (cell_num % cells_per_row) * cell_width + cell_width / 2;
+		y = BORDER_SIZE + (cell_num / cells_per_row) * cell_height + cell_height / 2;
+	}
+
+	public void GetCellSize (int cell_num, out int w, out int h)
+	{
+		// Trivial for now.
+		w = cell_width;
+		h = cell_height;
+	}
+
+
+	// Private utility methods.
 	public void UnselectAllCells ()
 	{
 		if (selected_cells.Count == 0)
@@ -243,7 +300,7 @@ public class IconView : Gtk.Layout {
 			SelectionChanged (this);
 	}
 
-	private bool CellIsSelected (int cell_num)
+	public bool CellIsSelected (int cell_num)
 	{
 		return selected_cells.ContainsKey (cell_num);
 	}
@@ -268,7 +325,7 @@ public class IconView : Gtk.Layout {
 
 	public void SelectAllCells ()
 	{
-		SelectCellRange (0, query.Photos.Length - 1);
+		SelectCellRange (0, collection.Photos.Length - 1);
 	}
 
 	private void SelectCellRange (int start, int end)
@@ -311,7 +368,15 @@ public class IconView : Gtk.Layout {
 
 
 	// Layout and drawing.
-
+	
+	// FIXME I can't find a c# wrapper for the C PANGO_PIXELS () macro
+	// So this Function is for that.
+	static int PangoPixels (int val)
+	{
+		return val >= 0 ? (val + 1024 / 2) / 1024 :
+			(val - 1024 / 2) / 1024;
+	}
+	
 	private void UpdateLayout ()
 	{
 		int available_width = Allocation.Width - 2 * BORDER_SIZE;
@@ -321,14 +386,18 @@ public class IconView : Gtk.Layout {
 
 		if (DisplayTags)
 			cell_height += TAG_ICON_SIZE + TAG_ICON_VSPACING;
+		
+		if (DisplayDates && this.Style != null) {
+			Pango.FontMetrics metrics = this.PangoContext.GetMetrics (this.Style.FontDescription, 
+										  Pango.Language.FromString ("en_US"));
+			cell_height += PangoPixels (metrics.Ascent + metrics.Descent);
+		}
 
-		cells_per_row = (int) (available_width / cell_width);
-		if (cells_per_row == 0)
-			cells_per_row = 1;
+		cells_per_row = Math.Max ((int) (available_width / cell_width), 1);
 
 		int num_thumbnails;
-		if (query != null)
-			num_thumbnails = query.Photos.Length;
+		if (collection != null)
+			num_thumbnails = collection.Photos.Length;
 		else
 			num_thumbnails = 0;
 
@@ -337,63 +406,136 @@ public class IconView : Gtk.Layout {
 			num_rows ++;
 
 		SetSize ((uint) Allocation.Width, (uint) (num_rows * cell_height + 2 * BORDER_SIZE));
+
+		Vadjustment.StepIncrement = cell_height;
+		Vadjustment.Change ();
 	}
 
+	System.Collections.Hashtable date_layouts = new Hashtable ();
 	// FIXME Cache the GCs?
-	private void DrawCell (int thumbnail_num, int x, int y)
+	private void DrawCell (int thumbnail_num, int x, int y, Gdk.Rectangle area)
 	{
 		Gdk.GC gc = new Gdk.GC (BinWindow);
 		gc.Copy (Style.ForegroundGC (StateType.Normal));
 		gc.SetLineAttributes (1, LineStyle.Solid, CapStyle.NotLast, JoinStyle.Round);
+		bool selected = CellIsSelected (thumbnail_num);
 
-		Photo photo = query.Photos [thumbnail_num];
+		Photo photo = collection.Photos [thumbnail_num];
 
 		string thumbnail_path = Thumbnail.PathForUri ("file://" + photo.DefaultVersionPath, ThumbnailSize.Large);
 		Pixbuf thumbnail = ThumbnailCache.Default.GetThumbnailForPath (thumbnail_path);
 
-		Gdk.Rectangle area = new Gdk.Rectangle (x, y, cell_width, cell_height);
-		Style.PaintBox (Style, BinWindow, StateType.Normal, ShadowType.Out, area, this, null, x, y, cell_width, cell_height);
+			
+		StateType cell_state = selected ? (HasFocus ? StateType.Selected :StateType.Active) : StateType.Normal;
 
-		if (HasFocus && thumbnail_num == focus_cell) {
-			Style.PaintFocus(Style, BinWindow, StateType.Normal, area, this, null, x + 3, y + 3, cell_width - 6, cell_height - 6);
+
+		Style.PaintFlatBox (Style, BinWindow, cell_state, 
+				    ShadowType.Out, area, this, "IconView", x, y, cell_width - 1, cell_height - 1);
+
+		if (HasFocus && thumbnail_num == FocusCell) {
+			Style.PaintFocus(Style, BinWindow, cell_state, area, 
+					 this, null, x + 3, y + 3, cell_width - 6, cell_height - 6);
 		}
 
-		if (thumbnail == null) {
-			pixbuf_loader.Request (thumbnail_path, thumbnail_num);
-		} else {
-			int width, height;
-			PixbufUtils.Fit (thumbnail, ThumbnailWidth, ThumbnailHeight, false, out width, out height);
+		Gdk.Rectangle image_area = new Gdk.Rectangle (x + CELL_BORDER_WIDTH, y + CELL_BORDER_WIDTH, cell_width - 2 * CELL_BORDER_WIDTH, cell_height - 2 * CELL_BORDER_WIDTH);
+		Gdk.Rectangle result = Rectangle.Zero;
 
-			Pixbuf temp_thumbnail;
-			if (width == thumbnail.Width)
-				temp_thumbnail = thumbnail;
+		int layout_width = 0;
+		int layout_height = 0;		
+
+		if (image_area.Intersect (area, out result)) {
+			int expansion = 0;
+			if (thumbnail_num == throb_cell) {
+				double t = throb_state / (double) (throb_state_max - 1);
+				double s;
+			if (selected)
+				s = Math.Cos (-2 * Math.PI * t);
 			else
-				temp_thumbnail = thumbnail.ScaleSimple (width, height, InterpType.Nearest);
-
-			int dest_x = (int) (x + (cell_width - width) / 2);
-			int dest_y;
-			if (DisplayTags)
-				dest_y = (int) (y + (cell_height - height - (TAG_ICON_SIZE + TAG_ICON_VSPACING)) / 2);
-			else
-				dest_y = (int) (y + (cell_height - height) / 2);
-
-			temp_thumbnail.RenderToDrawable (BinWindow, Style.WhiteGC,
-							 0, 0, dest_x, dest_y, width, height, RgbDither.None, 0, 0);
-
-			if (CellIsSelected (thumbnail_num)) {
-				Gdk.GC selection_gc = new Gdk.GC (BinWindow);
-
-				if (HasFocus)
-					selection_gc.Copy (Style.BackgroundGC (StateType.Selected));
-				else
-					selection_gc.Copy (Style.BackgroundGC (StateType.Active));
-
-				selection_gc.SetLineAttributes (SELECTION_THICKNESS, LineStyle.Solid, CapStyle.Butt, JoinStyle.Miter);
-
-				BinWindow.DrawRectangle (selection_gc, false,
-							 dest_x - SELECTION_THICKNESS, dest_y - SELECTION_THICKNESS,
-							 width + 2 * SELECTION_THICKNESS, height + 2 * SELECTION_THICKNESS);
+				s = 1 - Math.Cos (-2 * Math.PI * t);
+			
+			expansion = (int) (SELECTION_THICKNESS * s);
+			} else if (selected) {
+				expansion = SELECTION_THICKNESS;
 			}
+			
+			bool avoid_loader = false;
+			if (avoid_loader) {
+				if (thumbnail == null) {
+					//System.Console.WriteLine ("path = {0}", thumbnail_path);
+					thumbnail = new Pixbuf (thumbnail_path);
+					ThumbnailCache.Default.AddThumbnail (thumbnail_path, thumbnail);
+					thumbnail = new Pixbuf (thumbnail, 0, 0, thumbnail.Width, thumbnail.Height);
+				}
+
+				if (thumbnail == null)
+					thumbnail = PixbufUtils.ErrorPixbuf;
+			} else {
+				if (thumbnail == null) {
+					pixbuf_loader.Request (thumbnail_path, thumbnail_num);
+				}
+			}
+			
+			if (thumbnail != null){
+				int width, height;
+				PixbufUtils.Fit (thumbnail, ThumbnailWidth, ThumbnailHeight, true, out width, out height);
+				
+				int dest_x = (int) (x + (cell_width - width) / 2);
+				int dest_y;
+				
+				
+				dest_y = (int) y + ThumbnailHeight - height + CELL_BORDER_WIDTH;
+				
+				dest_x -= expansion;
+				dest_y -= expansion;		
+				width += 2 * expansion;
+				height += 2 * expansion;
+
+				Pixbuf temp_thumbnail;
+				if (width == thumbnail.Width) {
+					temp_thumbnail = thumbnail;
+				} else {
+					if (ThumbnailWidth > 64)
+						temp_thumbnail = thumbnail.ScaleSimple (width, height, InterpType.Bilinear);
+					else {
+						temp_thumbnail = thumbnail.ScaleSimple (width, height, InterpType.Nearest);
+					}
+				}
+				
+				Style.PaintShadow (Style, BinWindow, cell_state,
+						   ShadowType.Out, area, this, "IconView", dest_x - 1, dest_y - 1, width + 2, height + 2);			
+				temp_thumbnail.RenderToDrawable (BinWindow, Style.WhiteGC,
+								 0, 0, dest_x, dest_y, width, height, RgbDither.None, 0, 0);
+				
+				if (temp_thumbnail != thumbnail)
+					temp_thumbnail.Dispose ();
+				
+				thumbnail.Dispose ();
+			}
+		}
+			
+		if (DisplayDates) {
+			string date;
+			if (cell_width > 200) {
+				date = photo.Time.ToString ();
+			} else {
+				date = photo.Time.ToShortDateString ();
+			}
+
+			Pango.Layout layout = (Pango.Layout)date_layouts [date];
+			if (layout == null) {
+				layout = new Pango.Layout (this.PangoContext);
+				layout.SetText (date);
+				date_layouts [date] = layout;
+			}
+			
+			layout.GetPixelSize (out layout_width, out layout_height);
+
+			int layout_y = y + cell_height - CELL_BORDER_WIDTH - (DisplayTags ? TAG_ICON_SIZE : 0) - layout_height;
+			int layout_x = x + (cell_width - layout_width) / 2;
+
+			Style.PaintLayout (Style, BinWindow, cell_state,
+					   true, area, this, "IconView", layout_x, layout_y, layout);
+
 		}
 
 		if (DisplayTags) {
@@ -401,7 +543,7 @@ public class IconView : Gtk.Layout {
 
 			int tag_x, tag_y;
 
-			tag_x = x + CELL_BORDER_WIDTH;
+			tag_x = x + (cell_width - tags.Length * TAG_ICON_SIZE) / 2;
 			tag_y = y + cell_height - CELL_BORDER_WIDTH - TAG_ICON_SIZE;
 			foreach (Tag t in tags) {
 				Pixbuf icon = null;
@@ -428,21 +570,24 @@ public class IconView : Gtk.Layout {
 							      0, 0, tag_x, tag_y, TAG_ICON_SIZE, TAG_ICON_SIZE,
 							      RgbDither.None, 0, 0);
 				tag_x += TAG_ICON_SIZE + TAG_ICON_VSPACING;
+
+				if (scaled_icon != icon)
+					scaled_icon.Dispose ();
 			}
 		}
 	}
 
-	private void DrawAllCells (int x, int y, int width, int height)
+	private void DrawAllCells (Gdk.Rectangle area)
 	{
-		int start_cell_column = Math.Max ((x - BORDER_SIZE) / cell_width, 0);
-		int start_cell_row = Math.Max ((y - BORDER_SIZE) / cell_height, 0);
+		int start_cell_column = Math.Max ((area.X - BORDER_SIZE) / cell_width, 0);
+		int start_cell_row = Math.Max ((area.Y - BORDER_SIZE) / cell_height, 0);
 		int start_cell_num = start_cell_column + start_cell_row * cells_per_row;
 
 		int start_cell_x, cell_y;
 		GetCellPosition (start_cell_num, out start_cell_x, out cell_y);
 
-		int end_cell_column = Math.Max ((x + width - BORDER_SIZE) / cell_width, 0);
-		int end_cell_row = Math.Max ((y + height - BORDER_SIZE) / cell_height, 0);
+		int end_cell_column = Math.Max ((area.X + area.Width - BORDER_SIZE) / cell_width, 0);
+		int end_cell_row = Math.Max ((area.Y + area.Height - BORDER_SIZE) / cell_height, 0);
 
 		int num_rows = end_cell_row - start_cell_row + 1;
 		int num_cols = Math.Min (end_cell_column - start_cell_column + 1,
@@ -450,18 +595,23 @@ public class IconView : Gtk.Layout {
 
 		int i, cell_num;
 		for (i = 0, cell_num = start_cell_num;
-		     i < num_rows && cell_num < query.Photos.Length;
+		     i < num_rows && cell_num < collection.Photos.Length;
 		     i ++) {
 			int cell_x = start_cell_x;
 
-			for (int j = 0; j < num_cols && cell_num + j < query.Photos.Length; j ++) {
-				DrawCell (cell_num + j, cell_x, cell_y);
+			//Console.WriteLine ("Drawing row {0}", start_cell_row + i);
+			for (int j = 0; j < num_cols && cell_num + j < collection.Photos.Length; j ++) {
+				Gdk.Rectangle cell_bounds = CellBounds (cell_num + j);
+				if (area.Intersect (cell_bounds, out cell_bounds)) {
+					DrawCell (cell_num + j, cell_x, cell_y, cell_bounds);
+				}
 				cell_x += cell_width;
 			}
 
 			cell_y += cell_height;
 			cell_num += cells_per_row;
 		}
+
 	}
 
 	private void GetCellPosition (int cell_num, out int x, out int y)
@@ -487,7 +637,7 @@ public class IconView : Gtk.Layout {
 		if (y_offset == adjustment.Value)
 			return false;
 
-		int num_thumbnails = query.Photos.Length;
+		int num_thumbnails = collection.Photos.Length;
 		int num_rows, start;
 
 		if (y_offset < adjustment.Value) {
@@ -508,7 +658,7 @@ public class IconView : Gtk.Layout {
 			if (start + i >= num_thumbnails)
 				break;
 
-			Photo photo = query.Photos [start + i];
+			Photo photo = collection.Photos [start + i];
 			string thumbnail_path = Thumbnail.PathForUri ("file://" + photo.DefaultVersionPath, ThumbnailSize.Large);
 			pixbuf_loader.Cancel (thumbnail_path);
 		}
@@ -531,7 +681,46 @@ public class IconView : Gtk.Layout {
 			GLib.Source.Remove (scroll_on_idle_id);
 	}
 
-	private void ScrollTo (int cell_num)
+
+	//
+	// The throb interface
+	//
+	private uint throb_timer_id;
+	private int throb_cell = -1;
+	private int throb_state;
+	private const int throb_state_max = 40;
+	public void Throb (int cell_num)
+	{
+		throb_state = 0;
+		throb_cell = cell_num;
+		if (throb_timer_id == 0)
+			throb_timer_id = GLib.Timeout.Add ((39000/throb_state_max)/100,
+							   new GLib.TimeoutHandler (HandleThrobTimer));
+
+		InvalidateCell (cell_num);
+	}
+	
+	private void CancelThrob ()
+	{
+		if (throb_timer_id != 0)
+			GLib.Source.Remove (throb_timer_id);
+	}
+
+	private bool HandleThrobTimer () 
+	{
+		//Console.WriteLine ("throb out {1} {0}", throb_cell, 1 - Math.Cos (throb_state));
+
+		InvalidateCell (throb_cell);
+		if (throb_state++ < throb_state_max) {
+			return true;
+		} else {
+			throb_cell = -1;
+			throb_timer_id = 0;
+			return false;
+		}
+	}
+
+	public void ScrollTo (int cell_num)
 	{
 		Adjustment adjustment = Vadjustment;
 		int x;
@@ -544,7 +733,11 @@ public class IconView : Gtk.Layout {
 		else if (y + cell_height > adjustment.Value + adjustment.PageSize)
 			adjustment.Value = y + cell_height - adjustment.PageSize;
 		
-		adjustment.ChangeValue();
+#if USEIDLE
+		Scroll ();
+#else
+		adjustment.Change ();
+#endif		
 	}
 
 
@@ -557,19 +750,35 @@ public class IconView : Gtk.Layout {
 
 	private void HandlePixbufLoaded (PixbufLoader loader, string path, int order, Pixbuf result)
 	{
-		if (result == null)
-			result = ErrorPixbuf ();
+		if (result == null) {
+			result = PixbufUtils.ErrorPixbuf;
+			//
+			// ThumbnailCache Takes Ownership and calls Dispose
+			// so we need to copy the ErrorPixbuf
+			//
+			result = new Pixbuf (result, 0, 0, result.Width, result.Height);
+		}
 
+		//Console.WriteLine ("adding {0}", path);
 		ThumbnailCache.Default.AddThumbnail (path, result);
 		InvalidateCell (order);
 	}
+	
+	public Gdk.Rectangle CellBounds (int cell) 
+	{
+		Rectangle bounds;
+		GetCellPosition (cell, out bounds.X, out bounds.Y);
+		bounds.Width = cell_width;
+		bounds.Height = cell_height;
+		return bounds;
+	}
 
-	public void InvalidateCell (int order) {
-		Rectangle cell_area;
-		GetCellPosition (order, out cell_area.X, out cell_area.Y);
-		cell_area.Width = cell_width;
-		cell_area.Height = cell_height;
-
+	public void InvalidateCell (int order) 
+	{
+		Rectangle cell_area = CellBounds (order);
+		// FIXME where are we computing the bounds incorrectly
+		cell_area.Width -= 1;
+		cell_area.Height -= 1;
 		BinWindow.InvalidateRect (cell_area, false);
 	}
 			
@@ -579,15 +788,18 @@ public class IconView : Gtk.Layout {
 			args.Vadjustment.ValueChanged += new EventHandler (HandleAdjustmentValueChanged);
 	}
 
-	private void HandleSizeAllocated (object sender, SizeAllocatedArgs args)
+	protected override void OnSizeAllocated (Gdk.Rectangle allocation)
 	{
+		base.OnSizeAllocated (allocation);
 		UpdateLayout ();
 	}
 
-	private void HandleExposeEvent (object sender, ExposeEventArgs args)
+	protected override bool OnExposeEvent (Gdk.EventExpose args)
 	{
-		DrawAllCells (args.Event.Area.X, args.Event.Area.Y,
-			      args.Event.Area.Width, args.Event.Area.Height);
+		foreach (Rectangle area in args.Region.GetRectangles ()) {
+			DrawAllCells (area);
+		}
+		return base.OnExposeEvent (args);
 	}
 
  	private void HandleButtonPressEvent (object obj, ButtonPressEventArgs args)
@@ -598,6 +810,7 @@ public class IconView : Gtk.Layout {
 
 		if (cell_num < 0) {
 			args.RetVal = false;
+			UnselectAllCells ();
 			return;
 		}
 
@@ -612,28 +825,28 @@ public class IconView : Gtk.Layout {
 			return;
 
 		case EventType.ButtonPress:
-			if (args.Event.Button != 1)
-				return;
-
 			GrabFocus ();
 			if ((args.Event.State & ModifierType.ControlMask) != 0) {
 				ToggleCell (cell_num);
 			} else if ((args.Event.State & ModifierType.ShiftMask) != 0) {
-				SelectCellRange (focus_cell, cell_num);
-			} else {
+				SelectCellRange (FocusCell, cell_num);
+			} else if (!CellIsSelected (cell_num)) {
 				UnselectAllCells ();
 				SelectCell (cell_num);
 			}
 
-			Gdk.Pointer.Grab (BinWindow, false,
-					  EventMask.ButtonReleaseMask | EventMask.Button1MotionMask,
-					  null, null, args.Event.Time);
+			if (args.Event.Button == 3){
+				ContextMenu (args, cell_num);
+				return;
+			}
+			
+			if (args.Event.Button != 1)
+				return;
 
 			click_x = (int) args.Event.X;
 			click_y = (int) args.Event.Y;
 
-			focus_cell = cell_num;
-
+			FocusCell = cell_num;
 			return;
 
 		default:
@@ -642,92 +855,94 @@ public class IconView : Gtk.Layout {
 		}
 	}
 
-	private void HandleButtonReleaseEvent (object sender, ButtonReleaseEventArgs args)
+	void ContextMenu (ButtonPressEventArgs args, int cell_num)
 	{
-		Gdk.Pointer.Ungrab (args.Event.Time);
-		in_drag = false;
-	}
-
- 	private void HandleMotionNotifyEvent (object sender, MotionNotifyEventArgs args)
- 	{
-		if (in_drag)
-			return;
-
-		if (! Gtk.Drag.CheckThreshold (this, click_x, click_y, (int) args.Event.X, (int) args.Event.Y))
-			return;
-
-		TargetList target_list = new TargetList ();
-		in_drag = true;
+		PhotoPopup popup = new PhotoPopup ();
+		popup.Activate (args.Event);
 	}
 	
+	private void HandleButtonReleaseEvent (object sender, ButtonReleaseEventArgs args)
+	{
+		int cell_num = CellAtPosition ((int) args.Event.X, (int) args.Event.Y);
+
+		args.RetVal = true;
+
+		if (cell_num < 0) {
+			args.RetVal = false;
+			return;
+		}
+		
+		switch (args.Event.Type) {
+			case EventType.ButtonRelease:
+				if ((args.Event.State & ModifierType.ControlMask) == 0 &&
+				    (args.Event.State & ModifierType.ShiftMask  ) == 0 &&
+					(selected_cells.Count>1)) {
+					UnselectAllCells ();
+					SelectCell (cell_num);
+				}
+	
+				break;
+
+			default:
+				args.RetVal = false;
+				break;
+		}
+					
+	}
+
 	private void HandleKeyPressEvent (object sender, KeyPressEventArgs args)
 	{
 		int focus_old;
 		args.RetVal = true;
 		bool shift = ModifierType.ShiftMask == (args.Event.State & ModifierType.ShiftMask);
 		bool control = ModifierType.ControlMask == (args.Event.State & ModifierType.ControlMask);
-		
 
-		focus_old = focus_cell;
+		focus_old = FocusCell;
 		switch (args.Event.Key) {
 			case Gdk.Key.Down:
-				focus_cell += cells_per_row;
+				FocusCell += cells_per_row;
 				break;
 			case Gdk.Key.Left:
-				focus_cell--;
+				FocusCell--;
 				break;
 			case Gdk.Key.Right:
-				focus_cell++;
+				FocusCell++;
 				break;
 			case Gdk.Key.Up:
-				focus_cell -= cells_per_row;
+				FocusCell -= cells_per_row;
 				break;
 			case Gdk.Key.Home:
-				focus_cell = 0;
+				FocusCell = 0;
 				break;
 			case Gdk.Key.End:
-				focus_cell = query.Photos.Length - 1; 
+				FocusCell = collection.Photos.Length - 1; 
 				break;
-			/*
 			case Gdk.Key.space:
-				ToggleCell (focus_cell);
+				ToggleCell (FocusCell);
 				break;
-			*/
 			default:	
 				args.RetVal = false;
 				return;		
 		}
 		
-		if (focus_cell < 0 || focus_cell > query.Photos.Length - 1) {
-			focus_cell = focus_old;
+		if (FocusCell < 0 || FocusCell > collection.Photos.Length - 1) {
+			FocusCell = focus_old;
 			args.RetVal = false;
 		}	
 
 		if (shift) {
-			SelectCellRange (focus_old, focus_cell);
+			SelectCellRange (focus_old, FocusCell);
 		} else if (!control) {
 			UnselectAllCells ();
-			SelectCell (focus_cell);
+			SelectCell (FocusCell);
 		} 
 	
-		if  (focus_cell != focus_old) { 
-			InvalidateCell (focus_old);	
-			InvalidateCell (focus_cell);
-		}
-		
-		ScrollTo (focus_cell);
+		ScrollTo (FocusCell);
 	}
 
 	private void HandleDestroyEvent (object sender, DestroyEventArgs args)
 	{
 		CancelScroll ();
-	}
-
-
-	// DnD event handlers.
-
-	private void HandleDragDrop (object sender, DragDropArgs args)
-	{
-		Console.WriteLine ("HandleDragDrop");
+		CancelThrob ();
 	}
 }
